{
 "cells": [
  {
   "cell_type": "markdown",
   "id": "fad3bb2a",
   "metadata": {},
   "source": [
    "# syscalls in Go\n",
    "\n",
    "I was reviewing a PR to a project I know little about the other day, and one of the test was marked with `Skip` and a comment that it takes too long.\n",
    "This piqued my curiosity, so I took a deeper look.\n",
    "The function in question takes a library of valid characters and produces a random string of some length from that library.\n",
    "It is very straightforward, and would have gone completely unnoticed had it not been for a bit of developer paranoia.\n",
    "The author wrote a test invoking the function 1 million times, and it takes a handful of seconds.\n",
    "\"That's odd!\", I thought.\n",
    "This is basically string and number manipulation; even inefficient code tends to run quickly on modern machines.\n",
    "So began my journey under the hood of the `crypto/rand` package.\n",
    "\n",
    "Here is the original function:\n",
    "\n",
    "```go\n",
    "func genRandString(n int) string {\n",
    "        pool := []rune(SOME_ODD_CHARS)\n",
    "        b := make([]rune, n)\n",
    "        for i := range b {\n",
    "                r, err := rand.Int(rand.Reader, big.NewInt(int64(len(pool))))\n",
    "                if err != nil {\n",
    "                        log.Fatalf(\"encountered an error %v\", err)\n",
    "                }\n",
    "                b[i] = pool[r.Int64()]\n",
    "        }\n",
    "        return string(b)\n",
    "}\n",
    "```\n",
    "\n",
    "I was in a hurry and not paying attention.\n",
    "My original glance at the code had me thinking that it was searching the space of valid permutations.\n",
    "This was an ignorant thought, but it made me wonder if the necessary bytes could be read at once.\n",
    "As it turns out, there is!\n",
    "\n",
    "Before we get to that, let me explain what the previous function does; it wasn't obvious to me at first.\n",
    "- Initialize a slice of default bytes (value 0) of length `n`\n",
    "- For every byte in the slice\n",
    "    - Read a random int in the range of the library.\n",
    "    - Update this byte with the random value\n",
    "- Return the mutated slice\n",
    "\n",
    "I dug into the `crypto/rand` library a bit and found that you can read the random bytes in one call.\n",
    "Here is the updated function:\n",
    "\n",
    "```go\n",
    "func genRandStringV2(n int) string {\n",
    "        pool := []rune(SOME_ODD_CHARS)\n",
    "        b := make([]rune, n)\n",
    "        r := make([]byte, n)\n",
    "        rand.Read(r)\n",
    "        for i := range r {\n",
    "                b[i] = pool[int(r[i])%len(pool)]\n",
    "        }\n",
    "        return string(b)\n",
    "}\n",
    "```\n",
    "\n",
    "Time to test!\n",
    "This won't run on the playground because it relies on `time` and it takes too long.\n",
    "\n",
    "### Test code\n",
    "\n",
    "```go\n",
    "package main\n",
    "\n",
    "import (\n",
    "        \"crypto/rand\"\n",
    "        \"fmt\"\n",
    "        \"log\"\n",
    "        \"math/big\"\n",
    "        \"os\"\n",
    "        \"time\"\n",
    ")\n",
    "\n",
    "const SOME_ODD_CHARS string = \"1234567890abcdefghijklmnopqrstuvwxyzABCDEFGHIJKLMNOPQRSTUVWXYZ&*?()\"\n",
    "\n",
    "func genRandString(n int) string {\n",
    "        pool := []rune(SOME_ODD_CHARS)\n",
    "        b := make([]rune, n)\n",
    "        for i := range b {\n",
    "                r, err := rand.Int(rand.Reader, big.NewInt(int64(len(pool))))\n",
    "                if err != nil {\n",
    "                        log.Fatalf(\"encountered an error %v\", err)\n",
    "                }\n",
    "                b[i] = pool[r.Int64()]\n",
    "        }\n",
    "        return string(b)\n",
    "}\n",
    "\n",
    "func genRandStringV2(n int) string {\n",
    "        pool := []rune(SOME_ODD_CHARS)\n",
    "        b := make([]rune, n)\n",
    "        r := make([]byte, n)\n",
    "        rand.Read(r)\n",
    "        for i := range r {\n",
    "                b[i] = pool[int(r[i])%len(pool)]\n",
    "        }\n",
    "        return string(b)\n",
    "}\n",
    "\n",
    "func main() {\n",
    "        if len(os.Args) != 2 {\n",
    "                log.Fatalf(\"usage: script [v1, v2]\")\n",
    "        }\n",
    "        if os.Args[1] == \"v1\" {\n",
    "                start := time.Now()\n",
    "                for i := 0; i < 1000000; i++ {\n",
    "                        genRandString(10)\n",
    "                }\n",
    "                fmt.Printf(\"First version took %v\\n\", time.Since(start))\n",
    "        } else {\n",
    "                start := time.Now()\n",
    "                for i := 0; i < 1000000; i++ {\n",
    "                        genRandStringV2(10)\n",
    "                }\n",
    "                fmt.Printf(\"Second version took %v\\n\", time.Since(start))\n",
    "        }\n",
    "}\n",
    "```\n",
    "\n",
    "### Performance\n",
    "\n",
    "```shell\n",
    "❯ go run cmd/main.go v1\n",
    "First version took 5.958197642s\n",
    "\n",
    "❯ go run cmd/main.go v2\n",
    "Second version took 375.601598ms\n",
    "```\n",
    "\n",
    "Awesome, mission accomplished!\n",
    "Leave a note on the PR and walk away...\n",
    "\n",
    "As I was writing the note, I realized that I completely misread the original function in the first place.\n",
    "Functionally speaking, there is no difference between the two versions.\n",
    "Rolling a fair die `n` times is statistically no different than rolling `n` dice one time.\n",
    "In my naïve view of computing things, I saw bunch of stuff flying around on the heap, some CPU, and dash of magic.\n",
    "\n",
    "I sat back and thought about this for a bit.\n",
    "Cryptographic randomness requires a bit of entropy, so maybe the difference is side effects.\n",
    "Maybe there is some setup overhead in the `crypto` library.\n",
    "\n",
    "There is an [init function](https://cs.opensource.google/go/go/+/master:src/crypto/rand/rand_getentropy.go) in the library, but that is only called once.\n",
    "The folks in our internal chat had to draw me a picture before the light went off, but the basic idea was \"it's faster to read `n` bytes at once than to read 1 byte `n` times.\"\n",
    "Again, my head was stuck in the land of pure functions; why should this matter?!\n",
    "... unless there are side effects.\n",
    "\n",
    "The right idea finally made its way into my thick skull.\n",
    "I did some interweb searching and figured out how to use `strace`.\n",
    "I first searched for `entropy`, but that came up empty.\n",
    "Next, I tried `random`.\n",
    "Eureka!\n",
    "\n",
    "### Tracing syscalls\n",
    "\n",
    "```shell\n",
    "❯ strace -T ./main v1 2>&1 | rg \"random\" | wc -l\n",
    "2979598\n",
    "\n",
    "❯ strace -T ./main v2 2>&1 | rg \"random\" | wc -l\n",
    "432364\n",
    "```\n",
    "\n",
    "Anyone who does frequent systems programming probably stopped reading long ago.\n",
    "My domain is network automation, so the majority of my code is talking to slow network devices or making calls to APIs over the network which tend to be slow as well.\n",
    "In other words, this overhead is so bad that scripts and programs often take several seconds to run.\n",
    "What I learned from this is that syscalls are black boxes just like the network.\n",
    "You are making an API call and turning over control to another entity.\n",
    "Who knows what the kernel is going to do or when it will get around to doing it.\n",
    "\n",
    "I am glad I chased this rabbit down the hole because I feel like it was a valuable lesson.\n",
    "No, I don't need to prematurely optimize away syscalls in network-heavy code, but I do need to expand my view of the computing universe and how things work under the hood.\n",
    "Supply chains have been disrupted, inflation is crazy high, and companies are looking to cut costs at every turn.\n",
    "As developers, we should probably try to be a bit more efficient in the systems we create because economic laws may bite us long before [Moore's law](https://www.intel.com/content/www/us/en/history/virtual-vault/articles/moores-law.html) does!"
   ]
  }
 ],
 "metadata": {
  "kernelspec": {
   "display_name": "Python 3 (ipykernel)",
   "language": "python",
   "name": "python3"
  },
  "language_info": {
   "codemirror_mode": {
    "name": "ipython",
    "version": 3
   },
   "file_extension": ".py",
   "mimetype": "text/x-python",
   "name": "python",
   "nbconvert_exporter": "python",
   "pygments_lexer": "ipython3",
   "version": "3.10.10"
  },
  "nikola": {
   "category": "",
   "date": "2023-02-26 15:12:11-06:00",
   "description": "",
   "link": "",
   "slug": "syscalls-in-go",
   "tags": "go",
   "title": "Syscalls in Go",
   "type": "text"
  }
 },
 "nbformat": 4,
 "nbformat_minor": 5
}
